/*******************************************************************************
* Jojo's Patch: apply patch from jdiff to a binary file
*
* Copyright (C) 2002 Joris Heirbaut (joheirba@tijd.com)
*
* Author                Version Date       Modification
* --------------------- ------- -------    -----------------------
* Joris Heirbaut        v0.0    10-06-2002 hashed compare
* Joris Heirbaut        v0.1    20-06-2002 optimized esc-sequences & lengths
* Joris Heirbaut        v0.4c   09-01-2003 use seek for DEL-instruction
* Joris Heirbaut        v0.5    13-05-2003 do not count past EOF
* Joris Heirbaut	v0.6	13-05-2005 large-file support
*
* Licence
* -------
*
* This program is free software. Terms of the GNU General Public License apply.
*
* This program is distributed WITHOUT ANY WARRANTY, without even the implied
* warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
* See the GNU General Public License for more details.a
* 
* A copy of the GNU General Public License if found in the file "Licence.txt"
* deliverd along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* Parts or all of this source code may only be reused within other GNU free, open
* source, software.
* So if your project is not an open source project, you are non entitled to read
* any further below this line!
*
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include "jdiffos.h"

/*
 * Global settings (may be modified by commandline options) 
 */
int giVerbse = 0;       /* Verbose level 0=no, 1=normal, 2=high            */
FILE *stddbg;           /* Debug output to stddbg or stdout                */

/*******************************************************************************
* Input routines
*******************************************************************************/

int GetByt( lpFil )
FILE *lpFil ;
{ return getc(lpFil) ;
}

off_t GetInt( lpFil )
FILE *lpFil ;
{ off_t liVal ;

  liVal = getc(lpFil) ;
  if (liVal < 252)
    return liVal + 1 ;
  else if (liVal == 252)
    return 253 + getc(lpFil) ;
  else if (liVal == 253) {
    liVal = getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    return liVal ;
  }
  else if (liVal == 254) {
    liVal = getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    return liVal ;
  } else {
#ifdef _JDIFF_LARGEFILE
    liVal = getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
    liVal = (liVal << 8) + getc(lpFil) ;
#else
    fprintf(stderr, "64-bit length numbers not supported!");
    exit(6);
#endif
  }
}

/*******************************************************************************
* Patch function
*******************************************************************************
* Input stream consists of a series of
*   <op> (<data> || <len>)
* where
*   <op>   = <ESC> (<MOD>||<INS>||<DEL>||<EQL>)
*   <data> = <chr>||<ESC><ESC>
*   <chr>  = any byte different from <ESC><MOD><INS><DEL> or <EQL>
*   <ESC><ESC> yields one <ESC> byte
*******************************************************************************/
void jpatch ( asFilOrg, asFilPch, asFilOut )
FILE *asFilOrg ;
FILE *asFilPch ;
FILE *asFilOut ;
{
  int liInp ;         /* Current input from patch file          */
  int liOrg ;         /* Current input from data file           */
  int liOpr ;         /* Current operand                        */
  off_t liOff ;       /* Current operand's offset               */
  off_t liCnt ;       /* Counter                                */
  int lbChg=false ;   /* Changing operand?                      */
  int lbEsc=false ;   /* Non-operand escape char found?	    */

  liOpr = ESC ;
  while ((liInp = GetByt(asFilPch)) != EOF) {
    if (liInp == ESC) {
      liInp = GetByt(asFilPch); 
      switch(liInp) {
        case MOD: 
          liOpr = MOD;
          if (giVerbse == 2) {
            fprintf(stddbg, P8zd" "P8zd" MOD ...    \n", ftell(asFilOrg)+1, ftell(asFilOut)+1) ;
          }
          lbChg = true;
          break ;
          
        case INS: 
          liOpr = INS;
          if (giVerbse == 2) {
            fprintf(stddbg, P8zd" "P8zd" INS ...    \n", ftell(asFilOrg)+1, ftell(asFilOut)+1)   ;
          }
          lbChg = true;
          break ;

        case DEL: 
          liOpr = DEL; 
          liOff = GetInt(asFilPch); 
          if (giVerbse > 1) {
            fprintf(stddbg, P8zd" "P8zd" DEL "P8zd"\n", ftell(asFilOrg)+1, ftell(asFilOut)+1, liOff)  ;
          }

          if (fseek(asFilOrg, liOff, SEEK_CUR) != 0) {
            fprintf(stderr, "Could not position on original file (seek back %d).\n", liInp);
            exit(5);
          }
          lbChg = true;
          break ;

        case EQL: 
          liOpr = EQL; 
          liOff = GetInt(asFilPch); 
          if (giVerbse > 1) {
            fprintf(stddbg, P8zd" "P8zd" EQL "P8zd"\n", ftell(asFilOrg)+1, ftell(asFilOut)+1, liOff) ;
          }

          for (liCnt=0; liCnt < liOff; liCnt++) {
            liOrg = getc(asFilOrg) ;
            putc(liOrg, asFilOut) ;
          }
          lbChg = true;
          break ;

        case BKT:
          liOpr = BKT ;
          liOff = GetInt(asFilPch) ;
          if (giVerbse > 1) {
            fprintf(stddbg, P8zd" "P8zd" BKT "P8zd"\n", ftell(asFilOrg)+1, ftell(asFilOut)+1, liOff)   ;
          }

          if (fseek(asFilOrg, - liOff, SEEK_CUR) != 0) {
            fprintf(stderr, "Could not position on original file (seek back %d).\n", liInp);
            exit(5);
          }
          lbChg = true;
          break ;

        case ESC:
          if (giVerbse > 2) {
            fprintf(stddbg, P8zd" "P8zd" ESC ESC\n", ftell(asFilOrg)+1, ftell(asFilOut)+1) ;
          }
          break;

        default:
          if (giVerbse > 2) {
            fprintf(stddbg, P8zd" "P8zd" ESC XXX\n", ftell(asFilOrg)+1, ftell(asFilOut)+1) ;
          }
          lbEsc = true;
          break;
      }
    }
    if ( lbChg ) {
      lbChg = false ;
    } else {
      switch (liOpr) {
        case DEL: break;
        case EQL: break;
        case BKT: break;
        case MOD:
          if (lbEsc) {
            putc(ESC, asFilOut) ;
            liOrg = getc(asFilOrg) ;
            if (giVerbse > 2) {
              fprintf(stddbg, P8zd" "P8zd" MOD %3o %3o\n", ftell(asFilOrg), ftell(asFilOut), liOrg, liInp) ;
            }
          }

          putc(liInp, asFilOut) ;
          liOrg = getc(asFilOrg) ;
          if (giVerbse > 2) {
            fprintf(stddbg, P8zd" "P8zd" MOD %3o %3o\n", ftell(asFilOrg), ftell(asFilOut), liOrg, liInp)  ;
          }
          break ;

        case INS :
          if (lbEsc) {
            if (giVerbse > 2) {
              fprintf(stddbg, P8zd" "P8zd" INS     %3o\n", ftell(asFilOrg)+1, ftell(asFilOut)+1, ESC)  ; 
            }
            putc(ESC, asFilOut) ;
          }
  
          if (giVerbse > 2) {
            fprintf(stddbg, P8zd" "P8zd" INS     %3o\n", ftell(asFilOrg)+1, ftell(asFilOut)+1, liInp)  ; 
          }

          putc(liInp, asFilOut) ;
          break ;
      }
    } /* if lbChg */

    lbEsc = false ;
  } /* while */
}

/*******************************************************************************
* Main function
*******************************************************************************/
int main(int aiArgCnt, char *acArg[])
{
  char *lcFilNamOrg;
  char *lcFilNamPch;
  char *lcFilNamOut;

  FILE *lpFilOrg;
  FILE *lpFilPch;
  FILE *lpFilOut;

  int lbOptArgDne=false;
  int liOptArgCnt=0;
  char lcHlp='\0';

  /* Read options */
  stddbg = stderr;
  while (! lbOptArgDne && (aiArgCnt-1 > liOptArgCnt)) {
    liOptArgCnt++ ;
    if (strcmp(acArg[liOptArgCnt], "-v") == 0) {
      giVerbse = 1;
    } else if (strcmp(acArg[liOptArgCnt], "-vv") == 0) {
      giVerbse = 2;
    } else if (strcmp(acArg[liOptArgCnt], "-vvv") == 0) {
      giVerbse = 3;
    } else if (strcmp(acArg[liOptArgCnt], "-d") == 0) {
      stddbg = stdout;
    } else if (strcmp(acArg[liOptArgCnt], "-h") == 0) {
      lcHlp = 'h' ;
    } else {
      lbOptArgDne = true; 
      liOptArgCnt--;
    }
  }

  /* Output greetings */  
  if ((giVerbse>0) || (lcHlp == 'h') || (aiArgCnt - liOptArgCnt < 3)) {
    fprintf(stddbg, "JPATCH - Jojo's binary patch v0.2 (Beta) July 2005\n") ;
    fprintf(stddbg, "Copyright (C) 2002 Joris Heirbaut (joheirba@tijd.com)\n");
    fprintf(stddbg, "\n") ;
    fprintf(stddbg, "This program is free software. Terms of the GNU General Public License apply.\n") ;
    fprintf(stddbg, "\n") ;
    fprintf(stddbg, "This program is distributed WITHOUT ANY WARRANTY, without even the implied\n");
    fprintf(stddbg, "warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
    fprintf(stddbg, "See the GNU General Public License for more details.\n") ;
    fprintf(stddbg, "\n");
  }

  if ((aiArgCnt - liOptArgCnt < 3) || (lcHlp == 'h')) { 
    fprintf(stddbg, "Usage: jpatch [options] <original file> <patch file> [<output file>]\n") ;
    fprintf(stddbg, "  -v               Verbose (greeting, results and tips).\n");
    fprintf(stddbg, "  -vv              Verbose (debug info).\n");
    fprintf(stddbg, "  -vvv             Verbose (more debug info).\n");
    fprintf(stddbg, "  -h               Help (this text).\n");
    /*fprintf(stddbg, "  -l               Ascii patch file.\n");*/
    fprintf(stddbg, "Principles:\n");
    fprintf(stddbg, "  JPATCH reapplies a diff file, generated by jdiff, to the <original file>,\n") ;
    fprintf(stddbg, "  restoring the <new file>. For example, if jpatch is called as follows:\n");
    fprintf(stddbg, "    jdiff data01.tar data02.tar data02.dif\n");
    fprintf(stddbg, "  then data02.tar can be restored as follows:\n");
    fprintf(stddbg, "    jpatch data01.tar data02.dif data02.tar\n");
    fprintf(stddbg, "\n");
                     /******************************************************************************/
    exit(1);
  }

  /* Read filenames */
  lcFilNamOrg = acArg[liOptArgCnt + 1];
  lcFilNamPch = acArg[liOptArgCnt + 2];
  if (aiArgCnt > liOptArgCnt + 3)
    lcFilNamOut = acArg[liOptArgCnt + 3] ;
  else
    lcFilNamOut = "-" ;

  /* Open files */
  if ( strcmp(lcFilNamOrg, "-") == 0 )
    lpFilOrg = stdin ;
  else
    lpFilOrg = fopen(lcFilNamOrg, "rb") ;
  if ( lpFilOrg == NULL ) {
    fprintf(stddbg, "Could not open data file %s for reading.\n", lcFilNamOrg) ;
    exit(2);
  }

  if ( strcmp(lcFilNamPch, "-") == 0 )
    lpFilPch = stdin ;
  else 
    lpFilPch = fopen(lcFilNamPch, "rb") ;
  if ( lpFilPch == NULL ) {
    fprintf(stddbg, "Could not open patch file %s for reading.\n", lcFilNamPch) ;
    fclose(lpFilOrg);
    exit(3);
  }

  if ( strcmp(lcFilNamOut, "-") == 0 ) 
    lpFilOut = stdout ;
  else 
    lpFilOut = fopen(lcFilNamOut, "wb") ;
  if ( lpFilOut == NULL ) {
    fprintf(stddbg, "Could not open output file for writing.\n") ;
    fclose(lpFilOrg) ;
    fclose(lpFilPch) ;
    exit(4);
  }
 
  /* Go... */ 
  jpatch(lpFilOrg, lpFilPch, lpFilOut);

  /* Close files */
  fclose(lpFilOrg);
  fclose(lpFilPch);

  exit(0);
} 
